"""Based on https://github.com/utiasDSL/crazyflow/blob/d38bb5c75fe6624972ccc18d89789d3636cfd8cd/crazyflow/sim/integration.py."""

from __future__ import annotations

from typing import TYPE_CHECKING

from scipy.spatial.transform import Rotation as R

if TYPE_CHECKING:
    from drone_estimators._typing import Array  # To be changed to array_api_typing later
    from drone_estimators.structs import UKFData


def integrate_UKFData(state: UKFData, state_dot: UKFData) -> UKFData:
    """Integrates UKFData properly."""
    next_pos, next_quat, next_vel, next_ang_vel, next_rotor_vel = _integrate(
        state.pos,
        state.quat,
        state.vel,
        state.ang_vel,
        state_dot.pos,
        state_dot.quat,
        state_dot.vel,
        state_dot.ang_vel,
        state.dt,
        state.rotor_vel,
        state_dot.rotor_vel,
    )
    # TODO unit norm of quaternion!
    # TODO implement different integrator types later
    return state.replace(
        pos=next_pos, quat=next_quat, vel=next_vel, ang_vel=next_ang_vel, rotor_vel=next_rotor_vel
    )


def _integrate(
    pos: Array,
    quat: Array,
    vel: Array,
    ang_vel: Array,
    pos_dot: Array,
    quat_dot: Array,
    vel_dot: Array,
    ang_vel_dot: Array,
    dt: float,
    rotor_vel: Array | None = None,
    rotor_vel_dot: Array | None = None,
) -> Array:  # TODO is actually tuple
    """Integrate the dynamics forward in time.

    Args:
        pos: The position of the drone.
        quat: The orientation of the drone as a quaternion.
        vel: The velocity of the drone.
        ang_vel: The angular velocity of the drone.
        pos_dot: The derivative of the position of the drone.
        quat_dot: The derivative of the quaternion of the drone.
        vel_dot: The derivative of the velocity of the drone.
        ang_vel_dot: The derivative of the angular velocity of the drone.
        dt: The time step to integrate over.
        rotor_vel: The forces for the motors.
        rotor_vel_dot: The derivative of the motor forces.

    Returns:
        The next position, quaternion, velocity, and roll, pitch, and yaw rates of the drone.
    """
    next_pos = pos + pos_dot * dt
    next_quat = (R.from_quat(quat) * R.from_rotvec(ang_vel * dt)).as_quat()
    next_vel = vel + vel_dot * dt
    next_ang_vel = ang_vel + ang_vel_dot * dt
    next_rotor_vel = None
    if rotor_vel is not None:
        next_rotor_vel = rotor_vel + rotor_vel_dot * dt

    return next_pos, next_quat, next_vel, next_ang_vel, next_rotor_vel



"""This file contains the estimator node for ROS 2.

Essentially this file provides a wrapper for the estimators to be used with ROS data. The estimates get published to ROS as well.

TODO Subscribed and published topics...
"""

from __future__ import annotations

import argparse
import logging
import multiprocessing as mp
import os
import pickle
import signal
import time
from collections import defaultdict, deque
from pathlib import Path
from typing import TYPE_CHECKING

# Since motion_capture_tracking might be running on multiple PCs simultaneously,
# the same /tf topics get published multiple times to the network. This is
# unintended behavior. Setting it to LOCALHOST will block publishing to the
# network. We also set it here in the estimators s.t. they also don't publish
# to the network. Usually, the estimators get started after
# motion_capture_tracking, but in case the estimators get started first, we
# get consistent behavior.
# Note: All ros nodes started after this line will also only publish locally!
# Identical console command: ROS_AUTOMATIC_DISCOVERY_RANGE="LOCALHOST"
os.environ["ROS_AUTOMATIC_DISCOVERY_RANGE"] = "LOCALHOST"

import numpy as np
import rclpy
import toml

# Message types: https://docs.ros2.org/foxy/api/geometry_msgs/index-msg.html
from drone_models.core import load_params
from drone_models.transform import pwm2force
from geometry_msgs.msg import PoseStamped, TwistStamped, WrenchStamped
from munch import Munch, munchify
from rclpy.qos import QoSHistoryPolicy, QoSProfile, QoSReliabilityPolicy
from scipy.spatial.transform import Rotation as R
from std_msgs.msg import Float64MultiArray
from std_srvs.srv import Trigger
from tf2_msgs.msg import TFMessage
from visualization_msgs.msg import MarkerArray

from drone_estimators.estimator import KalmanFilter
from drone_estimators.estimator_legacy import StateEstimator
from drone_estimators.ros_nodes.ros2_utils import (
    append_measurement,
    append_state,
    create_array,
    create_marker_array,
    create_pose,
    create_twist,
    create_wrench,
    find_transform,
    tf2array,
)

if TYPE_CHECKING:
    from multiprocessing.sharedctypes import SynchronizedArray
    from multiprocessing.synchronize import Barrier, Event


class MPEstimator:
    """This class starts an estimator and the necessary subscribers and publishers.

    Note: Since all results are published individually, one might want to
    synchronize on the receiving end. For that see
    https://docs.ros.org/en/rolling/p/flex_sync/
    https://github.com/ros2/message_filters

    In the future, a custom message type might be advantageous. However,
    to keep it simple, we use standard message types.
    """

    def __init__(self, settings: Munch):
        """TODO."""
        self.settings = settings

        ctx = mp.get_context("spawn")
        self._shutdown = ctx.Event()
        self._publish_update = ctx.Event()
        startup = ctx.Barrier(3)  # Main process, _subscriber_loop, _publisher_loop

        # Logger setup
        self.logger = logging.getLogger("ESTIMATOR" + "_" + settings.drone_name)
        self.logger.setLevel(logging.INFO)
        # Create console handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        # Create formatter with [name][LEVEL] prefix
        formatter = logging.Formatter("[%(name)s][%(levelname)s] %(message)s")
        # Add formatter to handler
        console_handler.setFormatter(formatter)
        # Add handler to logger (only once)
        if not self.logger.hasHandlers():
            self.logger.addHandler(console_handler)
        # We allocate a buffer that is used to store the most recent tf message from the tf
        # subscriber. The message contains four fields: The first element is the number of messages
        # since the array was last cleared. This allows us to throw warnings in case the estimator
        # loop cannot keep up. The second field is the timestamp in s, the third is the position
        # (3d) and the fourth is the quaternion (4d)
        self._tf_msg_buffer = ctx.Array("d", [0.0] * (1 + 1 + 3 + 4))
        # Allocate the command subscriber buffer
        cmd_dim = 4  # May change in the future to be dynamic
        self._cmd_msg_buffer = ctx.Array("d", [0.0] * (1 + 1 + cmd_dim))
        # Estimated state publisher buffers
        self._pose_buffer = ctx.Array("d", [0.0] * 7)
        self._twist_buffer = ctx.Array("d", [0.0] * 6)
        self._forces_buffer = ctx.Array("d", [0.0] * 4)  # Motor forces, optional
        self._wrench_buffer = ctx.Array("d", [0.0] * 6)  # External wrench, optional

        args = (
            settings.drone_name,
            self._tf_msg_buffer,
            self._cmd_msg_buffer,
            startup,
            self._shutdown,
        )
        self._sub_process = ctx.Process(target=self._subscriber_loop, args=args)

        args = (
            settings.drone_name,
            self._pose_buffer,
            self._twist_buffer,
            self._forces_buffer,
            self._wrench_buffer,
            self._publish_update,
            startup,
            self._shutdown,
        )
        self._pub_process = ctx.Process(target=self._publisher_loop, args=args)

        self._sub_process.start()
        self._pub_process.start()
        startup.wait(10.0)
        self.logger.info("Subscriber and Publisher process started.")

    def _init_estimator(self):
        self.input_needed = False
        self.initial_observation = None

        self.time_stamp_last_prediction = 0
        self.time_stamp_last_correction = 0
        self.perf_timings = deque(maxlen=5000)

        self.frequency = self.settings.frequency  # Hz # TODO get from vicon frequency

        self.current_header = None
        self.current_state = None

        # This is for storing the data (DEBUG_SAVE_DATA)
        self.data_meas = defaultdict(list)  # {"time": [], "pos": [], "quat": [], "command": []}
        self.data_est = defaultdict(list)

        match self.settings.estimator_type:
            case "legacy":
                self.estimator = StateEstimator((0.0001, 0.007, 0.09, 0.005, 0.07))
                if (
                    self.settings.estimate_rotor_vel
                    or self.settings.estimate_dist_f
                    or self.settings.estimate_dist_t
                ):
                    self.logger.warning(
                        "Legacy estimator does not support force or torque estimation!"
                    )
            case "ukf":
                self.input_needed = True
                self.params = load_params(self.settings.dynamics_model, self.settings.drone_config)
                self.estimator = KalmanFilter(
                    dt=1 / self.frequency,
                    model=self.settings.dynamics_model,
                    config=self.settings.drone_config,
                    estimate_rotor_vel=self.settings.estimate_rotor_vel,
                    estimate_dist_f=self.settings.estimate_dist_f,
                    estimate_dist_t=self.settings.estimate_dist_t,
                )
            case _:
                raise NotImplementedError(
                    f"Estimator type {self.settings.estimator_type} not implemented."
                )

        # Initialization
        self.logger.info("Waiting for initial measurement.")
        while not self._shutdown.is_set():
            with self._tf_msg_buffer.get_lock():
                data = np.asarray(self._tf_msg_buffer, dtype=np.float64, copy=True)
                self._tf_msg_buffer[0] = 0
            n_tf_msg, tf_timestamp, pos, quat = data[0], data[1], data[2:5], data[5:]

            if n_tf_msg >= 1:
                self.time_stamp_last_prediction = tf_timestamp
                self.time_stamp_last_correction = tf_timestamp
                self.estimator.set_state(pos, quat)
                self.logger.info("Initialized pos and quat.")
                break

            time.sleep(0.5)

        self.logger.info(f"Started estimator (process {os.getpid()})")

    def run(self):
        """Main estimator loop."""
        self._init_estimator()  # done here such that errors can be raised properly

        k = 0
        global_time = time.perf_counter()

        try:
            # self.logger.info(f"{self.estimator.data.pos=}")
            # Estimation loop
            while not self._shutdown.is_set():
                loop_start_time = time.time()

                with self._tf_msg_buffer.get_lock():
                    data = np.asarray(self._tf_msg_buffer, dtype=np.float64, copy=True)
                    self._tf_msg_buffer[0] = 0
                n_tf_msg, tf_timestamp, pos, quat = data[0], data[1], data[2:5], data[5:]
                with self._cmd_msg_buffer.get_lock():
                    data = np.asarray(self._cmd_msg_buffer, dtype=np.float64, copy=True)
                    self._cmd_msg_buffer[0] = 0
                n_cmd_messages, cmd_timestep, cmd = data[0], data[1], data[2:]

                if cmd_timestep < tf_timestamp - 1 and cmd_timestep > 0 and self.input_needed:
                    self.logger.warning("Last command is older than 1s. Assuming zeros as input.")
                    self.estimator.set_input(np.array([0, 0, 0, 0]))

                if n_cmd_messages >= 1 and self.input_needed:
                    # The command is as it is sent to the drone, meaning for attitude interface:
                    # roll (deg), pitch (deg), yaw (deg), thrust (PWM)
                    # All the models run with rad and N, so we need to convert the RPYT command
                    cmd[..., -1] = pwm2force(
                        cmd[..., -1], self.params["thrust_max"] * 4, self.params["pwm_max"]
                    )
                    cmd[..., :-1] = np.deg2rad(cmd[..., :-1])
                    self.estimator.set_input(cmd)  # TODO # compare times?

                if n_tf_msg > 2:
                    self.logger.warning("Dropping tf messages because estimator loop can't keep up")
                    # TODO check for frequencies. If Vicon is running at higher frequency, of course estimator cant keep up

                if n_tf_msg >= 1:
                    dt = tf_timestamp - self.time_stamp_last_prediction
                    self.time_stamp_last_prediction = tf_timestamp

                    self.estimator.predict(dt)
                    self.estimator.correct(pos, quat)

                    # estimator_data = self.estimator.step(pos, quat, dt)

                time_stamp_now = time.time()
                dt = time_stamp_now - self.time_stamp_last_prediction
                self.time_stamp_last_prediction = time_stamp_now
                estimator_data = self.estimator.predict(dt)

                # Giving new estimate to publisher
                with self._pose_buffer.get_lock():
                    self._pose_buffer[:3] = estimator_data.pos
                    self._pose_buffer[3:] = estimator_data.quat
                with self._twist_buffer.get_lock():
                    self._twist_buffer[:3] = estimator_data.vel
                    self._twist_buffer[3:] = estimator_data.ang_vel
                if estimator_data.rotor_vel is not None:
                    with self._forces_buffer.get_lock():
                        self._forces_buffer[:] = estimator_data.rotor_vel
                if estimator_data.dist_f is not None:
                    with self._wrench_buffer.get_lock():
                        self._wrench_buffer[:3] = estimator_data.dist_f
                        if estimator_data.dist_t is not None:
                            self._wrench_buffer[3:] = estimator_data.dist_t
                self._publish_update.set()

                if self.settings.save_data:
                    append_state(self.data_est, time_stamp_now, estimator_data)
                    if n_tf_msg >= 1:
                        if n_cmd_messages >= 1 and self.input_needed:
                            append_measurement(self.data_meas, tf_timestamp, pos, quat, cmd)
                        if not self.input_needed:
                            append_measurement(self.data_meas, tf_timestamp, pos, quat, None)

                # if k % 100 == 0:
                #     self.logger.info(f"{estimator_data.dist_f=}")

                remaining = (
                    (1 / self.frequency) - (time.time() - loop_start_time) - 1.25 * 1e-4
                )  # TODO remove magic number, replace with "controller"
                if k % 1000 == 999:
                    self.logger.info(f"Freq: {k / (time.perf_counter() - global_time)}")
                k += 1
                if remaining > 0:
                    time.sleep(remaining)
        except KeyboardInterrupt:
            self._shutdown.set()

    @staticmethod
    def _subscriber_loop(
        drone_name: str,
        _tf_msg_buffer: SynchronizedArray,
        _cmd_msg_buffer: SynchronizedArray,
        startup: Barrier,
        shutdown: Event,
    ):
        rclpy.init()
        node = rclpy.create_node("estimator_sub_" + drone_name)
        qos_profile = QoSProfile(
            reliability=QoSReliabilityPolicy.BEST_EFFORT,
            history=QoSHistoryPolicy.KEEP_LAST,
            depth=1,
        )
        signal.signal(signal.SIGINT, lambda c, _: shutdown.set())  # Gracefully handle Ctrl-C

        # last_quat = np.array([0.0, 0.0, 0.0, 1.0])
        # calibration_quat = np.array([0.0, 0.0, 0.0, 1.0])
        last_rot = R.from_quat(np.array([0.0, 0.0, 0.0, 1.0]))
        calibration_rot = R.from_quat(np.array([0.0, 0.0, 0.0, 1.0]))

        def tf_callback(msg: TFMessage):
            tf = find_transform(msg.transforms, drone_name)
            if tf is None:
                node.get_logger().warning(
                    f"Drone {drone_name} could not have been found. Occluded?",
                    throttle_duration_sec=0.5,
                )
                return
            _, pos, quat = tf2array(tf)
            # We do not use the ros header time here because we need to establish an ordering
            # between observations and the model loop. We are not certain that the ros time stamp
            # is comparable to the system time. Therefore, we create a new timestamp using the os
            # time.
            # TODO: Subtract a constant time to account for [Vicon -> ros2 pub -> ros2 sub] delay.
            time_stamp = time.time()
            nonlocal last_rot
            last_rot = R.from_quat(quat)
            with _tf_msg_buffer.get_lock():
                _tf_msg_buffer[0] += 1
                _tf_msg_buffer[1] = time_stamp
                _tf_msg_buffer[2:5] = pos
                _tf_msg_buffer[5:9] = (calibration_rot.inv() * last_rot).as_quat()

        def cmd_callback(msg: Float64MultiArray):
            # The command is as it is sent to the drone, meaning for attitude interface:
            # roll (deg), pitch (deg), yaw (deg), thrust (PWM)
            with _cmd_msg_buffer.get_lock():
                _cmd_msg_buffer[0] += 1
                _cmd_msg_buffer[1] = time.time()
                _cmd_msg_buffer[2:] = msg.data

        def calibration_callback(
            request: Trigger.Request, response: Trigger.Response
        ) -> Trigger.Response:
            rpy = last_rot.as_euler("xyz", degrees=True)
            max_angle = 20  # degrees
            if np.any(rpy > max_angle):
                node.get_logger().warning("Calibration failed.")
                response.success = False
                response.message = "Pose could not be calibrated, deck tilted too much."
                return response

            node.get_logger().warning("Calibration successful.")
            nonlocal calibration_rot
            calibration_rot = last_rot
            response.success = True
            response.message = "Pose calibrated successfully."
            return response

        sub_tf = node.create_subscription(TFMessage, "/tf", tf_callback, qos_profile=qos_profile)
        sub_cmd = node.create_subscription(
            Float64MultiArray,
            f"/drones/{drone_name}/command",
            cmd_callback,
            qos_profile=qos_profile,
        )
        sub_calib = node.create_service(
            Trigger, f"/drones/{drone_name}/calibration", calibration_callback
        )
        startup.wait(10.0)  # Register this process as ready for startup barrier

        while not shutdown.is_set():
            rclpy.spin_once(node, timeout_sec=0.1)
        sub_tf.destroy()
        sub_cmd.destroy()
        sub_calib.destroy()
        node.destroy_node()

    @staticmethod
    def _publisher_loop(
        drone_name: str,
        pose_buffer: SynchronizedArray,
        twist_buffer: SynchronizedArray,
        forces_buffer: SynchronizedArray,
        wrench_buffer: SynchronizedArray,
        update: Event,
        startup: Barrier,
        shutdown: Event,
    ):
        rclpy.init()
        node = rclpy.create_node("estimator_sub_" + drone_name)
        # TODO check if pubs are actually needed?
        qos_profile = QoSProfile(
            reliability=QoSReliabilityPolicy.BEST_EFFORT,
            history=QoSHistoryPolicy.KEEP_LAST,
            depth=1,
        )
        # pos, quat
        pub_pose = node.create_publisher(
            PoseStamped, f"/drones/{drone_name}/estimate/pose", qos_profile=qos_profile
        )
        # vel, ang_vel
        pub_twist = node.create_publisher(
            TwistStamped, f"/drones/{drone_name}/estimate/twist", qos_profile=qos_profile
        )
        # f_motors
        pub_forces = node.create_publisher(
            Float64MultiArray, f"/drones/{drone_name}/estimate/forces", qos_profile=qos_profile
        )
        # f_dis, t_dis
        pub_wrench = node.create_publisher(
            WrenchStamped, f"/drones/{drone_name}/estimate/wrench", qos_profile=qos_profile
        )
        # marker (only for rviz)
        pub_markers = node.create_publisher(
            MarkerArray, f"/drones/{drone_name}/estimate/marker_array", qos_profile=qos_profile
        )
        signal.signal(signal.SIGINT, lambda c, _: shutdown.set())  # Gracefully handle Ctrl-C
        startup.wait(10.0)  # Register this process as ready for startup barrier

        # TODO clear arrays???
        while not shutdown.is_set():
            if not update.wait(0.5):
                continue
            time_stamp = time.time()
            update.clear()

            pose = np.asarray(pose_buffer, dtype=np.float64, copy=True)
            pose_stamped = create_pose(time_stamp, drone_name, pose[:3], pose[3:])
            pub_pose.publish(pose_stamped)

            twist = np.asarray(twist_buffer, dtype=np.float64, copy=True)
            twist_stamped = create_twist(time_stamp, drone_name, twist[:3], twist[3:])
            pub_twist.publish(twist_stamped)

            forces = np.asarray(forces_buffer, dtype=np.float64, copy=True)
            # This type doesn't have a stamp!
            forces_array = create_array(time_stamp, drone_name, forces)
            pub_forces.publish(forces_array)

            wrench = np.asarray(wrench_buffer, dtype=np.float64, copy=True)
            wrench_stamped = create_wrench(time_stamp, drone_name, wrench[:3], wrench[3:])
            pub_wrench.publish(wrench_stamped)

            markers = create_marker_array(
                time_stamp,
                drone_name,
                pose[:3],
                pose[3:],
                twist[:3],
                twist[3:],
                wrench[:3],
                wrench[3:],
            )
            pub_markers.publish(markers)

        pub_pose.destroy()
        pub_twist.destroy()
        pub_forces.destroy()
        pub_wrench.destroy()
        node.destroy_node()

    def close(self):
        """TODO."""
        self.logger.info(f"Estimator {self.settings.drone_name} shutdown")
        self._shutdown.set()
        self._sub_process.join()
        self._pub_process.join()

        if self.settings.save_data:
            self.logger.info("Saving data...")
            filename = f"data_{self.settings.drone_name}_"
            info = f"{self.settings.estimator_type}"
            if self.settings.estimator_type != "legacy":
                info = info + f"_{self.settings.dynamics_model}"
            with open(filename + info + ".pkl", "wb") as f:
                pickle.dump(self.data_est, f)
            with open(filename + "measurement" + ".pkl", "wb") as f:
                pickle.dump(self.data_meas, f)


def launch_estimators(estimators: dict):
    """TODO."""
    processes = []
    seen_drone_names = []
    ctx = mp.get_context("spawn")
    shutdown = ctx.Event()

    try:
        for k, settings in estimators.items():
            name = settings.drone_name
            if name in seen_drone_names:
                print(f"[ESTIMATOR_{name}]  Estimator for {name} already existing. Check settings")
                continue
            seen_drone_names.append(name)
            # not sure if daemon should be True or False
            p = ctx.Process(target=launch_node, args=(settings, shutdown))
            processes.append(p)
            p.start()

        try:
            while True:
                time.sleep(10.0)
        except KeyboardInterrupt:
            print("\nKeyboard interrupt received. Terminating nodes...")

    finally:
        shutdown.set()

        for p in processes:
            p.join(timeout=2)

        for p in processes:
            if p.is_alive():
                print(f"Force terminating process {p.pid}")
                p.terminate()
                p.join()

        if rclpy.ok():
            rclpy.shutdown()
        print("All nodes terminated.")


def launch_node(settings: Munch, stop_event: Event):
    """TODO."""
    rclpy.init()
    estimator = MPEstimator(settings)
    try:
        estimator.run()
    finally:
        estimator.close()


if __name__ == "__main__":
    np.set_printoptions(linewidth=400, precision=3)  # TODO remove
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--settings", default=None, help="Path to settings file in CWD")
    parser.add_argument("--drone_name", default=None, help="Overwrite drone_name in estimator1")
    args = parser.parse_args()

    if args.settings is None:
        path = Path(__file__).parents[1] / "ros_nodes/estimators.toml"
    else:
        path = args.settings
    with open(path, "r") as f:
        estimators = munchify(toml.load(f))
    if args.drone_name is not None:
        estimators.estimator1.drone_name = args.drone_name  # overwrite drone_name

    # Add debug to each estimator (if not already in place)
    for key, val in estimators.items():
        if not key.startswith("estimator"):
            continue
        for global_key, global_val in estimators.get("global", {}).items():
            if global_key not in val:
                val[global_key] = global_val

    estimators = munchify({k: v for k, v in estimators.items() if k.startswith("estimator")})

    launch_estimators(estimators)



"""This file contains some util functions for ROS 2 nodes."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from geometry_msgs.msg import Point, PoseStamped, TransformStamped, TwistStamped, WrenchStamped
from scipy.spatial.transform import Rotation as R
from std_msgs.msg import Float64MultiArray
from visualization_msgs.msg import Marker, MarkerArray

if TYPE_CHECKING:
    from collections import defaultdict

    from std_msgs.msg import Header

    from drone_estimators._typing import Array  # To be changed to array_api_typing later
    from drone_estimators.structs import UKFData


def find_transform(
    transforms: list[TransformStamped], child_frame_id: str
) -> TransformStamped | None:
    """Finds the transform of a given child frame."""
    return next((tf for tf in transforms if tf.child_frame_id == child_frame_id), None)


def header2sec(header: Header) -> float:
    """Extracts the current time in seconds from a message header."""
    return float(header.stamp.sec + header.stamp.nanosec * 1e-9)


def tf2array(tf: TransformStamped) -> tuple[Header, Array, Array]:
    """Converts a Transform into numpy arrays."""
    pos = np.array(
        [tf.transform.translation.x, tf.transform.translation.y, tf.transform.translation.z]
    )

    quat = np.array(
        [
            tf.transform.rotation.x,
            tf.transform.rotation.y,
            tf.transform.rotation.z,
            tf.transform.rotation.w,
        ]
    )

    return tf.header, pos, quat


def time2sec_nsec(t: float) -> tuple[int, int]:
    """Converts time as float into time as sec and nsec for ROS."""
    sec = int(t)
    nsec = int(1e9 * (t - sec))
    return sec, nsec


def create_pose(time_stamp: float, drone: str, pos: Array, quat: Array) -> PoseStamped:
    """Creates a PoseStamped from position and orientation."""
    pose = PoseStamped()
    sec, nsec = time2sec_nsec(time_stamp)
    pose.header.stamp.sec = sec
    pose.header.stamp.nanosec = nsec
    pose.header.frame_id = "world"
    pose.pose.position.x = pos[0]
    pose.pose.position.y = pos[1]
    pose.pose.position.z = pos[2]
    pose.pose.orientation.x = quat[0]
    pose.pose.orientation.y = quat[1]
    pose.pose.orientation.z = quat[2]
    pose.pose.orientation.w = quat[3]

    return pose


def create_twist(time_stamp: float, drone: str, vel: Array, ang_vel: Array) -> TwistStamped:
    """Creates a TwistStamped based on velocity and angular velocity."""
    twist = TwistStamped()
    sec, nsec = time2sec_nsec(time_stamp)
    twist.header.stamp.sec = sec
    twist.header.stamp.nanosec = nsec
    twist.header.frame_id = drone
    twist.twist.linear.x = vel[0]
    twist.twist.linear.y = vel[1]
    twist.twist.linear.z = vel[2]
    twist.twist.angular.x = ang_vel[0]
    twist.twist.angular.y = ang_vel[1]
    twist.twist.angular.z = ang_vel[2]

    return twist


def create_array(time_stamp: float, drone: str, data: Array | None) -> Float64MultiArray:
    """Creates a generic Float64Array with the length of data.

    If data is None, an array zeros of length 4 is created.
    """
    if data is not None:
        array = Float64MultiArray()
        array.data = list(data)
        return array
    else:
        array = Float64MultiArray()
        array.data = [0.0, 0.0, 0.0, 0.0]
        return array


def create_wrench(
    time_stamp: float, drone: str, force: Array | None, torque: Array | None
) -> WrenchStamped:
    """Creates a WrenchStamped based on force and torque.

    If force or torque is None, it gets set to zeros.
    """
    wrench = WrenchStamped()
    sec, nsec = time2sec_nsec(time_stamp)
    wrench.header.stamp.sec = sec
    wrench.header.stamp.nanosec = nsec
    wrench.header.frame_id = drone
    if force is not None:
        wrench.wrench.force.x = force[0]
        wrench.wrench.force.y = force[1]
        wrench.wrench.force.z = force[2]
    else:
        wrench.wrench.force.x = 0.0
        wrench.wrench.force.y = 0.0
        wrench.wrench.force.z = 0.0
    if torque is not None:
        wrench.wrench.torque.x = 0.0
        wrench.wrench.torque.y = 0.0
        wrench.wrench.torque.z = 0.0

    return wrench


def make_arrow_marker_from_vector(
    id: int,
    ns: str,
    start_pos: Array,
    vector: Array,
    color_rgba: Array = [1, 1, 1, 1],
    scale: float = 1.0,
) -> Marker:
    """TODO."""
    marker = Marker()
    marker.id = id
    marker.ns = ns
    marker.header.frame_id = "world"
    marker.type = Marker.ARROW
    marker.action = Marker.ADD

    # Arrow from start_pos to start_pos + vector * scale
    start = Point()
    start.x = start_pos[0]
    start.y = start_pos[1]
    start.z = start_pos[2]

    end = Point()
    end.x = start.x + vector[0] * scale
    end.y = start.y + vector[1] * scale
    end.z = start.z + vector[2] * scale

    marker.points = [start, end]

    marker.scale.x = 0.01  # shaft diameter
    marker.scale.y = 0.05  # head diameter
    marker.scale.z = 0.05  # head length

    marker.color.r = color_rgba[0]
    marker.color.g = color_rgba[1]
    marker.color.b = color_rgba[2]
    marker.color.a = color_rgba[3]
    if np.linalg.norm(vector) < 0.1:
        marker.color.a = 0

    return marker


def create_marker_array(
    time_stamp: float,
    drone: str,
    pos: Array,
    quat: Array,
    vel: Array,
    ang_vel: Array,
    force: Array | None,
    torque: Array | None,
) -> MarkerArray:
    """Creates a MarkerArray for visualization in RViz."""
    marker_array = MarkerArray()
    # print(f"{quat=}")

    ### Drone marker
    marker = Marker()
    sec, nsec = time2sec_nsec(time_stamp)
    marker.header.stamp.sec = sec
    marker.header.stamp.nanosec = nsec
    marker.header.frame_id = "world"
    marker.pose.position.x = pos[0]
    marker.pose.position.y = pos[1]
    marker.pose.position.z = pos[2]
    marker.pose.orientation.x = quat[0]
    marker.pose.orientation.y = quat[1]
    marker.pose.orientation.z = quat[2]
    marker.pose.orientation.w = quat[3]

    marker.ns = "drone"
    marker.id = 0
    marker.type = Marker.MESH_RESOURCE
    marker.action = Marker.ADD
    marker.scale.x = marker.scale.y = marker.scale.z = 1.0
    marker.color.r = marker.color.g = marker.color.b = 1.0
    marker.color.a = 1.0
    marker.mesh_resource = "package://motion_capture_tracking/meshes/crazyflie2.dae"
    marker.mesh_use_embedded_materials = True
    marker_array.markers.append(marker)

    ### Velocity Marker
    marker_array.markers.append(
        make_arrow_marker_from_vector(1, "twist", pos, vel, [1.0, 0.0, 0.0, 1.0])
    )

    ### Angular Velocity Marker
    rot = R.from_quat(quat)
    ang_vel = rot.apply(ang_vel)  # Rotating the vector same as the done to be displayed correctly
    marker_array.markers.append(
        make_arrow_marker_from_vector(2, "twist", pos, ang_vel, [1.0, 1.0, 0.0, 1.0])
    )

    ### Force Marker
    marker_array.markers.append(
        make_arrow_marker_from_vector(3, "wrench", pos, force * 5, [0.0, 0.0, 1.0, 1.0])
    )

    ### Torque Marker
    marker_array.markers.append(
        make_arrow_marker_from_vector(4, "wrench", pos, torque * 5, [1.0, 0.0, 1.0, 1.0])
    )

    return marker_array


def append_state(data: defaultdict[str, list], time: float, state: UKFData):
    """Appends each states data to the corresponding of list in the dictionary."""
    data["time"].append(time)
    data["pos"].append(state.pos)
    data["quat"].append(state.quat)
    data["vel"].append(state.vel)
    data["ang_vel"].append(state.ang_vel)
    data["covariance"].append(state.covariance)
    if state.forces_motor is not None:
        data["forces_motor"].append(state.forces_motor)
    else:
        data["forces_motor"] = []
    if state.forces_dist is not None:
        data["forces_dist"].append(state.forces_dist)
    else:
        data["forces_dist"] = []
    if state.torques_dist is not None:
        data["torques_dist"].append(state.torques_dist)
    else:
        data["torques_dist"] = []


def append_measurement(
    data: defaultdict[str, list], time: float, pos: Array, quat: Array, command: Array | None = None
):
    """Appends each measurment data to the corresponding of list in the dictionary."""
    data["time"].append(time)
    data["pos"].append(pos)
    data["quat"].append(quat)
    if command is None:
        command = np.zeros_like(quat)
    data["command"].append(command)



"""TODO."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from flax.struct import Callable, dataclass

if TYPE_CHECKING:
    from drone_estimators._typing import Array  # To be changed to array_api_typing later


@dataclass
class UKFData:
    """TODO."""

    pos: Array
    quat: Array
    vel: Array
    ang_vel: Array
    rotor_vel: Array | None
    dist_f: Array | None
    dist_t: Array | None
    covariance: Array  # Covariance matrix

    sigmas_f: Array
    sigmas_h: Array

    u: Array  # input
    z: Array  # measurement
    dt: float

    @classmethod
    def create_empty(
        cls,
        rotor_vel: bool = False,
        dist_f: bool = False,
        dist_t: bool = False,
        dim_u: int = 4,
        dim_z: int = 7,
    ) -> UKFData:
        """TODO."""
        pos = np.zeros(3)
        quat = np.array([0, 0, 0, 1])
        vel = np.zeros(3)
        ang_vel = np.zeros(3)
        dim_x = 13
        if rotor_vel:
            rotor_vel = np.zeros(4)
            dim_x = dim_x + 4
        else:
            rotor_vel = None
        if dist_f:
            dist_f = np.zeros(3)
            dim_x = dim_x + 3
        else:
            dist_f = None
        if dist_t:
            dist_t = np.zeros(3)
            dim_x = dim_x + 3
        else:
            dist_t = None

        covariance = np.eye(dim_x)

        sigmas_f = np.zeros((2 * dim_x + 1, dim_x))
        sigmas_h = np.zeros((2 * dim_x + 1, dim_z))

        u = np.zeros(dim_u)  # input
        z = np.zeros(dim_z)  # measurement
        dt = 1

        return cls(
            pos,
            quat,
            vel,
            ang_vel,
            rotor_vel,
            dist_f,
            dist_t,
            covariance,
            sigmas_f,
            sigmas_h,
            u,
            z,
            dt,
        )

    @classmethod
    def create(
        cls,
        pos: Array,
        quat: Array,
        vel: Array,
        ang_vel: Array,
        rotor_vel: Array | None = None,
        dist_f: Array | None = None,
        dist_t: Array | None = None,
    ) -> UKFData:
        """TODO."""
        dim_x = 13
        if rotor_vel is None:
            dim_x = dim_x + 4
        if dist_f is None:
            dim_x = dim_x + 3
        if dist_t is None:
            dim_x = dim_x + 3

        covariance = np.eye(dim_x)

        sigmas_f = np.zeros((2 * dim_x + 1, dim_x))
        sigmas_h = np.zeros((2 * dim_x + 1, 7))

        u = np.zeros(4)  # input
        z = np.zeros(7)  # measurement
        dt = 1

        return cls(
            pos,
            quat,
            vel,
            ang_vel,
            rotor_vel,
            dist_f,
            dist_t,
            covariance,
            sigmas_f,
            sigmas_h,
            u,
            z,
            dt,
        )

    @classmethod
    def as_state_array(cls, data: UKFData) -> Array:
        """Returns the state as an array."""
        xp = data.pos.__array_namespace__()
        x = xp.concat((data.pos, data.quat, data.vel, data.ang_vel), axis=-1)
        if data.rotor_vel is not None:
            x = xp.concat((x, data.rotor_vel), axis=-1)
        if data.dist_f is not None:
            x = xp.concat((x, data.dist_f), axis=-1)
        if data.dist_t is not None:
            x = xp.concat((x, data.dist_t), axis=-1)
        return x

    @classmethod
    def from_state_array(cls, data: UKFData, array: Array) -> UKFData:
        """Updates data in the given structure based on a given array."""
        pos = array[..., 0:3]
        quat = array[..., 3:7]
        vel = array[..., 7:10]
        ang_vel = array[..., 10:13]
        idx = 13
        if data.rotor_vel is not None:
            rotor_vel = array[..., idx : idx + 4]
            idx = idx + 4
        else:
            rotor_vel = None
        if data.dist_f is not None:
            dist_f = array[..., idx : idx + 3]
            idx = idx + 3
        else:
            dist_f = None
        if data.dist_t is not None:
            dist_t = array[..., idx : idx + 3]
            idx = idx + 3
        else:
            dist_t = None

        return data.replace(
            pos=pos,
            quat=quat,
            vel=vel,
            ang_vel=ang_vel,
            rotor_vel=rotor_vel,
            dist_f=dist_f,
            dist_t=dist_t,
        )

    @classmethod
    def get_state_dim(cls, data: UKFData) -> int:
        """Returns the dimension of the state."""
        dim_x = 13
        if data.rotor_vel is not None:
            dim_x = dim_x + 4
        if data.dist_f is not None:
            dim_x = dim_x + 3
        if data.dist_t is not None:
            dim_x = dim_x + 3
        return dim_x


@dataclass
class UKFSettings:
    """TODO."""

    SPsettings: SigmaPointsSettings
    Q: Array
    R: Array
    fx: Callable[
        [Array, Array, Array, Array, Array, Array, Array | None, Array | None],
        tuple[Array, Array, Array, Array, Array | None],
    ]
    hx: Callable[[Array, Array, Array, Array, Array, Array, Array | None, Array | None], Array]

    @classmethod
    def create(
        cls,
        SPsettings: SigmaPointsSettings,
        Q: Array,
        R: Array,
        fx: Callable[
            [Array, Array, Array, Array, Array, Array, Array | None, Array | None],
            tuple[Array, Array, Array, Array, Array | None],
        ],
        hx: Callable[[Array, Array, Array, Array, Array, Array, Array | None, Array | None], Array],
    ) -> UKFSettings:
        """TODO."""
        return cls(SPsettings, Q, R, fx, hx)


@dataclass
class SigmaPointsSettings:
    """TODO."""

    n: int
    alpha: float
    beta: float
    kappa: float
    lambda_: float
    Wc: Array
    Wm: Array

    @classmethod
    def create(cls, n: int, alpha: float, beta: float, kappa: float = 0.0) -> SigmaPointsSettings:
        """TODO."""
        lambda_ = alpha**2 * (n + kappa) - n
        c = 0.5 / (n + lambda_)
        Wc0 = np.array([lambda_ / (n + lambda_) + (1 - alpha**2 + beta)])
        Wm0 = np.array([lambda_ / (n + lambda_)])
        Wc = np.full(2 * n, c)
        Wm = np.full(2 * n, c)
        Wc = np.concat((Wc0, Wc))
        Wm = np.concat((Wm0, Wm))

        return cls(n, alpha, beta, kappa, lambda_, Wc, Wm)



"""Testing the selfimplemented rotations against scipy rotations."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from drone_models import available_models, model_features
from drone_models.drones import available_drones

from drone_estimators.estimator import KalmanFilter
from drone_estimators.utils.dynamics import dynamics_function

if TYPE_CHECKING:
    from typing import Callable


@pytest.mark.unit
def test_placeholder():
    """Placeholder test."""
    pass


@pytest.mark.unit
@pytest.mark.parametrize("model_name, model", available_models.items())
@pytest.mark.parametrize("drone_type", available_drones)
@pytest.mark.unit
def test_model_loading(model_name: str, model: Callable, drone_type: str):
    """Tests if the models for the kalman filters can be imported."""
    dynamics_function(model_name, drone_type)


# TODO test if the whole filterpy chain is jitable


@pytest.mark.unit
@pytest.mark.parametrize("model_name, model", available_models.items())
@pytest.mark.parametrize("drone_type", available_drones)
@pytest.mark.unit
def test_kalman(model_name: str, model: Callable, drone_type: str):
    """Tests if the Kalman filter can be imported and stepped."""
    supports_dynamics = model_features(dynamics_function(model_name, drone_type))["rotor_dynamics"]
    kf = KalmanFilter(1 / 200, model_name, drone_type, estimate_rotor_vel=supports_dynamics)

    kf.predict(1 / 240, np.array([0.0, 0.0, 0.0, 0.5]))

    kf.correct(np.array([0.0, 0.0, 0.0]), np.array([0.0, 0.0, 0.0, 1.0]))



"""This file contains modified methods and classes taken from the filterpy library.

Copyright 2015 Roger R Labbe Jr.

FilterPy library.
http://github.com/rlabbe/filterpy

Documentation at:
https://filterpy.readthedocs.org

Supporting book at:
https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python

This is licensed under an MIT license. See the readme.MD file
for more information.
"""

from __future__ import absolute_import, annotations, division, print_function

from typing import TYPE_CHECKING

# import jax
import numpy as np
from scipy.linalg import block_diag

from drone_estimators.integration import integrate_UKFData
from drone_estimators.structs import UKFData, UKFSettings

if TYPE_CHECKING:
    from drone_estimators._typing import Array  # To be changed to array_api_typing later

# rotor_vel: Array | None = None,
#     dist_f: Array | None = None,
#     dist_t


# @jax.jit
def ukf_predict_correct(data: UKFData, settings: UKFSettings) -> UKFData:
    """TODO."""
    xp = data.pos.__array_namespace__()
    #### Predict
    # Calculate sigma points
    # TODO special sigma points for quaternions!
    sigmas = ukf_calculate_sigma_points(data, settings)
    data_sigmas = UKFData.from_state_array(data, sigmas)

    # Pass sigma points through dynamics
    pos_dot, quat_dot, vel_dot, ang_vel_dot, rotor_vel_dot = settings.fx(
        pos=data_sigmas.pos,
        quat=data_sigmas.quat,
        vel=data_sigmas.vel,
        ang_vel=data_sigmas.ang_vel,
        cmd=data.u,
        rotor_vel=data_sigmas.rotor_vel,
        dist_f=data_sigmas.dist_f,
        dist_t=data_sigmas.dist_t,
    )
    data_sigmas_dot = UKFData.create(pos_dot, quat_dot, vel_dot, ang_vel_dot, rotor_vel_dot)

    # Integrate dynamics if continuous
    data_sigmas_f = integrate_UKFData(data_sigmas, data_sigmas_dot)
    sigmas_f = UKFData.as_state_array(data_sigmas_f)
    data = data.replace(sigmas_f=sigmas_f)

    # Pass prior sigmas through measurment function h(x,u,dt) to get measurement sigmas
    sigmas_h = settings.hx(
        pos=data_sigmas_f.pos,
        quat=data_sigmas_f.quat,
        vel=data_sigmas_f.vel,
        ang_vel=data_sigmas_f.ang_vel,
        cmd=data.u,
        rotor_vel=data_sigmas_f.rotor_vel,
        dist_f=data_sigmas_f.dist_f,
        dist_t=data_sigmas_f.dist_t,
    )
    data = data.replace(sigmas_h=sigmas_h)

    # Compute prior with unscented transform
    x, P = ukf_unscented_transform(
        sigmas_f, settings.SPsettings.Wm, settings.SPsettings.Wc, settings.Q
    )

    # save prior
    data = UKFData.from_state_array(data, x)
    data = data.replace(covariance=P)

    #### Correct
    # Pass mean and covariance of prediction through unscented transform
    zp, S = ukf_unscented_transform(
        data.sigmas_h, settings.SPsettings.Wm, settings.SPsettings.Wc, settings.R
    )
    # SI = xp.linalg.inv(data.S)
    # data = data.replace(S=S, SI=SI)
    # data = data.replace(S=S)

    # compute cross variance
    Pxz = ukf_cross_variance(
        UKFData.as_state_array(data), zp, data.sigmas_f, data.sigmas_h, settings.SPsettings.Wc
    )
    # K = xp.dot(Pxz, data.SI)       # Kalman gain
    # K @ S = Pxz => K = Pxz @ S^-1 => or: S.T @ K.T = Pxz.T
    K = xp.linalg.solve(S.T, Pxz.T).T
    y = xp.subtract(data.z, zp)  # residual
    # data = data.replace(K=K, y=y)

    # Update Gaussian state estimate (x, P)
    x = UKFData.as_state_array(data) + xp.dot(K, y)
    # print(f"P prior = {xp.diag(P)}")
    # print(f"P prior = \n{P}")
    # Added identity for numerical stability
    P = data.covariance - xp.dot(K, xp.dot(S, K.T))  # + xp.eye(P.shape[0]) * 1e-9
    # print(f"P post = {xp.diag(P)}")
    # print(f"P post = \n{P}")

    # Save posterior
    data = UKFData.from_state_array(data, x)
    data = data.replace(covariance=P)

    return data


# @jax.jit
def ukf_predict(data: UKFData, settings: UKFSettings) -> UKFData:
    """TODO."""
    # xp = data.pos.__array_namespace__()
    #### Predict
    # Calculate sigma pointstf.transform.rotation.x,
    # TODO special sigma points for quaternions!
    sigmas = ukf_calculate_sigma_points(data, settings)
    data_sigmas = UKFData.from_state_array(data, sigmas)

    # Pass sigma points through dynamics
    pos_dot, quat_dot, vel_dot, ang_vel_dot, rotor_vel_dot = settings.fx(
        pos=data_sigmas.pos,
        quat=data_sigmas.quat,
        vel=data_sigmas.vel,
        ang_vel=data_sigmas.ang_vel,
        cmd=data.u,
        rotor_vel=data_sigmas.rotor_vel,
        dist_f=data_sigmas.dist_f,
        dist_t=data_sigmas.dist_t,
    )
    # print(f"{pos_dot=}")
    data_sigmas_dot = UKFData.create(pos_dot, quat_dot, vel_dot, ang_vel_dot, rotor_vel_dot)

    # Integrate dynamics if continuous
    data_sigmas_f = integrate_UKFData(data_sigmas, data_sigmas_dot)
    sigmas_f = UKFData.as_state_array(data_sigmas_f)
    data = data.replace(sigmas_f=sigmas_f)

    # Pass prior sigmas through measurment function h(x,u,dt) to get measurement sigmas
    sigmas_h = settings.hx(
        pos=data_sigmas_f.pos,
        quat=data_sigmas_f.quat,
        vel=data_sigmas_f.vel,
        ang_vel=data_sigmas_f.ang_vel,
        cmd=data.u,
        rotor_vel=data_sigmas_f.rotor_vel,
        dist_f=data_sigmas_f.dist_f,
        dist_t=data_sigmas_f.dist_t,
    )
    data = data.replace(sigmas_h=sigmas_h)

    # Compute prior with unscented transform
    x, P = ukf_unscented_transform(
        sigmas_f, settings.SPsettings.Wm, settings.SPsettings.Wc, settings.Q
    )

    # save prior
    data = UKFData.from_state_array(data, x)
    data = data.replace(covariance=P)

    return data


# @jax.jit
def ukf_correct(data: UKFData, settings: UKFSettings) -> UKFData:
    """TODO."""
    xp = data.covariance.__array_namespace__()
    # Pass mean and covariance of prediction through unscented transform
    zp, S = ukf_unscented_transform(
        data.sigmas_h, settings.SPsettings.Wm, settings.SPsettings.Wc, settings.R
    )
    # SI = xp.linalg.inv(data.S)
    # data = data.replace(S=S, SI=SI)
    # data = data.replace(S=S)

    # compute cross variance
    Pxz = ukf_cross_variance(
        UKFData.as_state_array(data), zp, data.sigmas_f, data.sigmas_h, settings.SPsettings.Wc
    )
    # K = xp.dot(Pxz, data.SI)       # Kalman gain
    # K @ S = Pxz => K = Pxz @ S^-1 => or: S.T @ K.T = Pxz.T
    K = xp.linalg.solve(S.T, Pxz.T).T
    y = xp.subtract(data.z, zp)  # residual
    # data = data.replace(K=K, y=y)

    # Update Gaussian state estimate (x, P)
    x = UKFData.as_state_array(data) + xp.dot(K, y)
    # print(f"P prior = {xp.diag(P)}")
    # print(f"P prior = \n{P}")
    # Added identity for numerical stability
    P = data.covariance - xp.dot(K, xp.dot(S, K.T))  # + xp.eye(P.shape[0]) * 1e-9
    # print(f"P post = {xp.diag(P)}")
    # print(f"P post = \n{P}")

    # Save posterior
    data = UKFData.from_state_array(data, x)
    data = data.replace(covariance=P)

    return data


def ukf_calculate_sigma_points(data: UKFData, settings: UKFSettings) -> Array:
    """TODO."""
    xp = data.pos.__array_namespace__()
    P = data.covariance
    # Adding some very small identity part for numerical stability
    # Note: Higher values make the system more stable for the cost of more noise!
    P = P + xp.eye(P.shape[0]) * 1e-12
    U = xp.linalg.cholesky((settings.SPsettings.lambda_ + settings.SPsettings.n) * P, upper=True)

    state_array = UKFData.as_state_array(data)
    sigma_center = state_array
    # TODO for the quaternions use more sophisitcated approach to keep length 1!
    # Mainly: rotate in tangent space
    sigma_pos = xp.subtract(state_array, -U)
    sigma_neg = xp.subtract(state_array, U)
    return xp.vstack((sigma_center, sigma_pos, sigma_neg))


def ukf_unscented_transform(
    sigmas: Array, Wm: Array, Wc: Array, noise_cov: Array = None
) -> tuple[Array, Array]:
    """TODO."""
    xp = sigmas.__array_namespace__()
    x = xp.dot(Wm, sigmas)

    # new covariance is the sum of the outer product
    # of the residuals times the weights
    y = sigmas - x[None, :]
    P = xp.dot(y.T, xp.dot(xp.diag(Wc), y))

    if noise_cov is not None:
        P = P + noise_cov

    return (x, P)


def ukf_cross_variance(x: Array, z: Array, sigmas_f: Array, sigmas_h: Array, Wc: Array) -> Array:
    """Compute cross variance of the state `x` and measurement `z`."""
    xp = x.__array_namespace__()
    # The slicing brings Wc in the correct shape to be broadcast
    # The einsum as set up here takes the outer product of all the stacked vectors
    Pxz = Wc[:, None, None] * xp.einsum(
        "bi,bj->bij", xp.subtract(sigmas_f, x), xp.subtract(sigmas_h, z)
    )
    Pxz = xp.sum(Pxz, axis=0)
    return Pxz


def order_by_derivative(Q: Array, dim: int, block_size: int) -> Array:
    """TODO."""
    xp = Q.__array_namespace__()

    N = dim * block_size

    D = xp.zeros((N, N))

    Q = xp.array(Q)
    for i, x in enumerate(Q.ravel()):
        f = xp.eye(block_size) * x

        ix, iy = (i // dim) * block_size, (i % dim) * block_size
        D[ix : ix + block_size, iy : iy + block_size] = f

    return D


def Q_discrete_white_noise(
    dim: int, dt: float = 1.0, var: float = 1.0, block_size: int = 1, order_by_dim: bool = True
) -> Array:
    """TODO."""
    if not (dim == 2 or dim == 3 or dim == 4):
        raise ValueError("dim must be between 2 and 4")

    if dim == 2:
        Q = [[0.25 * dt**4, 0.5 * dt**3], [0.5 * dt**3, dt**2]]
    elif dim == 3:
        Q = [
            [0.25 * dt**4, 0.5 * dt**3, 0.5 * dt**2],
            [0.5 * dt**3, dt**2, dt],
            [0.5 * dt**2, dt, 1],
        ]
    else:
        Q = [
            [(dt**6) / 36, (dt**5) / 12, (dt**4) / 6, (dt**3) / 6],
            [(dt**5) / 12, (dt**4) / 4, (dt**3) / 2, (dt**2) / 2],
            [(dt**4) / 6, (dt**3) / 2, dt**2, dt],
            [(dt**3) / 6, (dt**2) / 2, dt, 1.0],
        ]

    if order_by_dim:
        return block_diag(*[Q] * block_size) * var
    return order_by_derivative(np.array(Q), dim, block_size) * var



"""This file implements some wrappers to get the dynamics from 'drone-models'."""

from __future__ import annotations

import importlib
from typing import TYPE_CHECKING

from drone_models.core import parametrize

if TYPE_CHECKING:
    from collections.abc import Callable

    from drone_estimators._typing import Array  # To be changed to array_api_typing later


def dynamics_function(model: str, config: str) -> Callable:
    """Imports and return the correct dynamics function from the drone-models."""
    module = importlib.import_module(f"drone_models.{model}.model")
    fn = getattr(module, "dynamics")
    return parametrize(fn, config)  # parametrize to remove model parameters from function


def observation_function(
    pos: Array,
    quat: Array,
    vel: Array,
    ang_vel: Array,
    cmd: Array,
    rotor_vel: Array | None = None,
    dist_f: Array | None = None,
    dist_t: Array | None = None,
) -> Array:
    """Return the observable part of the state.

    This is basically not necessary, since we always get position and orientation
    from Vicon. However, for sake of completeness, this observation function is added.
    """
    xp = pos.__array_namespace__()
    return xp.concat((pos, quat), axis=-1)



"""This file is to be remove later as soon as a proper typing is available by the official array-api."""

from typing import Any, TypeAlias

import numpy.typing as npt

Array: TypeAlias = Any  # To be changed to array_api_typing later
ArrayLike: TypeAlias = Array | npt.ArrayLike



"""A simple smoothing estimator.

This estimator is originally from Chris McKinnon.
It was adapted to work with similarly to the new estimators for backwards compatability.
"""

from __future__ import absolute_import, annotations, division, print_function

import math
from threading import Lock
from typing import TYPE_CHECKING

import numpy as np
import transforms3d as tf

from drone_estimators.structs import UKFData

if TYPE_CHECKING:
    from drone_estimators._typing import Array  # To be changed to array_api_typing later

# from drone_estimators.quaternions import apply_omega_to_quat, global_to_body, omega_from_quat_quat


class StateEstimator(object):
    """Vicon state estimation and filtering.

    Parameters
    ----------
    filter_parameters : sequence of floats
        The 4 tuning parameters for the Kalman filter
    """

    def __init__(self, filter_parameters: tuple):
        """TODO."""
        # Lock for access to state
        # Makes sure the service does not conflict with normal updates
        self.state_access_lock = Lock()

        # Read in the parameters
        (
            self.tau_est_trans,
            self.tau_est_trans_dot,
            self.tau_est_trans_dot_dot,
            self.tau_est_rot,
            self.tau_est_rot_dot,
        ) = filter_parameters

        # Initialize the state variables
        self.pos = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.vel = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.acc = np.array([0.0, 0.0, 0.0], dtype=np.float64)

        # Initialize rotations (quaternions).
        self.quat = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
        self.omega_g = np.array([0.0, 0.0, 0.0], dtype=np.float64)

        # Initialize measurements
        self.pos_meas = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.vel_meas = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.acc_meas = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.quat_meas = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
        self.omega_g_meas = np.array([0.0, 0.0, 0.0], dtype=np.float64)

        # Initialze old measurements
        self.pos_old = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.vel_old = np.array([0.0, 0.0, 0.0], dtype=np.float64)
        self.quat_old = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)

        # Correction to align vicon frame with body frame
        self.quat_corr = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)

        # Initialize time..at first time step trusts measurements to 100%
        self.time = 0.0
        self.time_meas = 0.0
        self.dt = 0.0

    @property
    def rpy(self):
        """Convert quaternion to roll, pitch, yaw."""
        raise NotImplementedError()
        # return tf.euler_from_quaternion(self.quat)
        # return R.from_quat(self.quat).as_euler("xyz")

    @property
    def omega_b(self) -> Array:
        """Return the body angular velocity."""
        return global_to_body(self.quat, self.omega_g)

    def step(self, pos: Array, quat: Array, dt: float, command: Array | None = None) -> UKFData:
        """This function is not part of the legacy estimator and only for compatability."""
        self.dt = dt
        # the transforms3d library used here uses scalar first, so we need to make it scalar first
        quat = np.roll(quat, 1)
        self.get_new_measurement(pos, quat)
        self.prior_update()
        self.measurement_update()
        # outside, we use scalar last, so we need to roll back
        return UKFData.create(
            self.pos, np.array(np.roll(self.quat, -1)).astype(float), self.vel, self.omega_b
        )

    def predict(self, dt: float, command: Array | None = None) -> UKFData:
        """This function is not part of the legacy estimator and only for compatability."""
        # Since the legacy estimator doesn't inherently support the prediction/correction
        # form of a Kalman filter, we only step the time in the prediction step. In the
        # correction step, we use the accumulated time step size to correct with the
        # correct dt.
        self.dt += dt
        return UKFData.create(
            self.pos, np.array(np.roll(self.quat, -1)).astype(float), self.vel, self.omega_b
        )

    def correct(self, pos: Array, quat: Array, command: Array | None = None) -> UKFData:
        """This function is not part of the legacy estimator and only for compatability."""
        # Since the legacy estimator doesn't inherently support the prediction/correction
        # form of a Kalman filter, we only step the time in the prediction step. In the
        # correction step, we use the accumulated time step size to correct with the
        # correct dt.
        data = self.step(pos, quat, self.dt, command)
        self.dt = 0
        return data

    def set_input(self, command: Array):
        """This function is not part of the legacy estimator and only for compatability."""
        pass

    def set_state(self, pos: Array, quat: Array):
        """This function is not part of the legacy estimator and only for compatability."""
        self.pos = pos
        self.quat = np.roll(quat, 1)

    def get_new_measurement(self, position: Array, quaternion: Array):
        """Get a new measurement of position and orientation.

        Parameters
        ----------
        time: float
            The measurement time in seconds.
        position: ndarray
        quaternion: ndarray
        """
        # Record the time at which the state was determined
        # self.time_meas = time

        # Calculate time difference, update time
        # self.dt = self.time_meas - self.time

        # Get the translational position from VICON
        self.pos_meas = position

        # Get the rotational position in the form of a quaternion
        self.quat_meas = quaternion

        # Apply correction
        # self.quat_meas = tf.quaternion_multiply(self.quat_meas, self.quat_corr)
        self.quat_meas = tf.quaternions.qmult(self.quat_meas, self.quat_corr)

        # Two quaternions for every rotation, make sure we take
        # the one that is consistent with previous measurements
        if np.dot(self.quat_old, self.quat_meas) < 0.0:
            self.quat_meas = -self.quat_meas

        # Don't compute finite difference for impossibly small time differences
        if self.dt <= 1e-15:
            return

        # Numeric derivatives: Comput velocities
        self.vel_meas = (self.pos_meas - self.pos_old) / self.dt

        # Numeric derivatives: Compute accelerations
        self.acc_meas = (self.vel_meas - self.vel_old) / self.dt

        # Numeric derivatives: Compute angular velocity
        self.omega_g_meas = omega_from_quat_quat(self.quat_old, self.quat_meas, self.dt)

        # Update old measurements (make a copy)
        self.pos_old[:] = self.pos_meas
        self.vel_old[:] = self.vel_meas
        self.quat_old[:] = self.quat_meas

    def prior_update(self):
        """Predict future states using a double integrator model."""
        # Acceleration and angular velocity are assumed constant
        # Update position and orientation
        self.pos += self.dt * self.vel + 0.5 * self.dt * self.dt * self.acc
        self.vel += self.dt * self.acc
        self.quat = apply_omega_to_quat(self.quat, self.omega_g, self.dt)

    def measurement_update(self):
        """Update state estimate with measurements."""
        # NOTE: Use raw position and quaternion - no low pass filter
        # Calculate current Kalman filter gains
        c1 = math.exp(-self.dt / self.tau_est_trans)
        c2 = math.exp(-self.dt / self.tau_est_trans_dot)
        c3 = math.exp(-self.dt / self.tau_est_trans_dot_dot)

        d1 = math.exp(-self.dt / self.tau_est_rot)
        d2 = math.exp(-self.dt / self.tau_est_rot_dot)

        # Wait while locked, then lock itself
        with self.state_access_lock:
            # Measurement updates
            self.pos = (1.0 - c1) * self.pos_meas + c1 * self.pos
            self.vel = (1.0 - c2) * self.vel_meas + c2 * self.vel
            self.acc = (1.0 - c3) * self.acc_meas + c3 * self.acc

            self.quat = (1.0 - d1) * self.quat_meas + d1 * self.quat
            self.omega_g = (1.0 - d2) * self.omega_g_meas + d2 * self.omega_g

            # Make sure that numerical errors don't pile up
            # self.quat /= tf.vector_norm(self.quat)
            self.quat /= tf.quaternions.qnorm(self.quat)

            self.time = self.time_meas


def omega_from_quat_quat(q1: Array, q2: Array, dt: float) -> Array:
    """Convert two quaternions and the time difference to angular velocity.

    Parameters:
    -----------
    q1: quaternion
        The old quaternion
    q2: quaternion
        The new quaternion
    dt: float
        The time difference

    Returns:
    --------
    omega_g: ndarray
        The angular velocity in global coordinates
    """
    if tf.quaternions.qnorm(q1 - q2) < 1e-8:
        # linearly interpolate
        # the quaternion does not stay on unit sphere -> only for very small
        # rotations!

        # dq/dt
        dq = (q2 - q1) / dt

        # From Diebel: Representing Atitude, 6.6, quaternions are defined
        # differently there: [w, x, y, z] instead of [x, y, z, w]!
        omega = np.array([0.0, 0.0, 0.0], dtype=np.float64)

        omega[0] = 2.0 * (q2[0] * dq[1] - q2[3] * dq[2] + q2[2] * dq[3] - q2[1] * dq[0])
        omega[1] = 2.0 * (q2[3] * dq[1] + q2[0] * dq[2] - q2[1] * dq[3] - q2[2] * dq[0])
        omega[2] = 2.0 * (-q2[2] * dq[1] + q2[1] * dq[2] + q2[0] * dq[3] - q2[3] * dq[0])

        return omega
    else:
        # This function becomes numerically unstable for q1-q2 --> 0

        # Find rotation from q1 to q2
        # unit quaternion -> conjugate is the same as inverse
        # q2 = r * q1 --> r = q2 * inv(q1)
        r = tf.quaternions.qmult(q2, tf.quaternions.qconjugate(q1))
        # r = tf.quaternion_multiply(q2, tf.quaternion_conjugate(q1))
        r /= tf.quaternions.qnorm(r)

        # Angle of rotation
        # angle = 2.0 * math.acos(r[3])
        angle = 2.0 * math.acos(r[0])

        # acos gives value in [0,pi], ensure that we take the short path
        # (e.g. rotate by -pi/2 rather than 3pi/2)
        if angle > math.pi:
            angle -= 2.0 * math.pi

        # angular velocity = angle / dt
        # axis of rotation corresponds to r[:3]
        # return angle / dt * r[:3] / tf.quaternions.qnorm(r[:3])
        return angle / dt * r[1:] / tf.quaternions.qnorm(r[1:])


def apply_omega_to_quat(q: Array, omega: Array, dt: float) -> Array:
    """Convert a quaternion q and apply the angular velocity omega to it over dt.

    Parameters:
    -----------
    q: quaternion
    omega: ndarray
        angular velocity
    dt: float
        time difference

    Returns:
    --------
    quaternion
        The quaternion of the orientation after rotation with omega for dt
        seconds.
    """
    # rotation angle around each axis
    w = omega * dt

    # only rotate if the angle we rotate through is actually significant
    if tf.quaternions.qnorm(w) < np.finfo(float).eps * 4.0:
        return q

    # quaternion corresponding to this rotation
    # w = 0 is not a problem because numpy is awesome
    r = tf.quaternions.axangle2quat(w, tf.quaternions.qnorm(w))
    # r = tf.quaternion_about_axis(np.linalg.norm(w), w)

    # return the rotated quaternion closest to original
    return tf.quaternions.qmult(r, q)


def global_to_body(q: Array, vec: Array) -> Array:
    """Convert a vector from global to body coordinates.

    Parameters:
    -----------
    q: quaternion
        The rotation quaternion
    vec: ndarray
        The vector in global coordinates

    Returns:
    vec: ndarray
        The vector in body coordinates
    """
    # tf.quaternion_matrix(q)[:3,:3] is a homogenous rotation matrix that
    # rotates a vector by q
    # tf.quaternion_matrix(q)[:3,:3] is rot. matrix from body to global frame
    # its transpose is the trafo matrix from global to body
    # that matrix is multiplied by omega
    return np.dot(tf.quaternions.quat2mat(q).transpose(), vec)


def body_to_global(q: Array, vec: Array) -> Array:
    """Convert a vector from global to body coordinates.

    Parameters:
    -----------
    q: quaternion
        The rotation quaternion
    vec: ndarray
        The vector in body coordinates

    Returns:
    vec: ndarray
        The vector in global coordinates
    """
    # tf.quaternion_matrix(q)[:3,:3] is a homogenous rotation matrix that
    # rotates a vector by q
    # tf.quaternion_matrix(q)[:3,:3] is the matrix from body to global frame
    # that matrix is multiplied by omega
    return np.dot(tf.quaternions.quat2mat(q), vec)
    # return np.dot(tf.quaternion_matrix(q)[:3, :3], vec)



"""TODO."""

from __future__ import absolute_import, annotations, division, print_function

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from drone_estimators.filterpy import (
    Q_discrete_white_noise,
    ukf_correct,
    ukf_predict,
    ukf_predict_correct,
)
from drone_estimators.structs import SigmaPointsSettings, UKFData, UKFSettings
from drone_estimators.utils.dynamics import dynamics_function, observation_function

if TYPE_CHECKING:
    from drone_estimators._typing import Array  # To be changed to array_api_typing later


class Estimator(ABC):
    """Base class for estimator implementations."""

    def __init__(self, state_dim: int, input_dim: int, obs_dim: int, dt: float):
        """Initialize basic parameters.

        Args:
            state_dim: Dimensionality of the systems states, e.g., x of f(x,u)
            input_dim: Dimensionality of the input to the dynamics, e.g., u of f(x,u)
            obs_dim: Dimensionality of the observations, e.g., y
            dt: Time step between callings.
        """
        self._state_dim = state_dim
        self._obs_dim = obs_dim
        self._input_dim = input_dim
        self._dt = dt
        self._state = np.zeros(self._state_dim)
        self._input = np.zeros(self._input_dim)

    def reset(self):
        """Reset the noise to its initial state."""
        self._state = np.zeros(self._state_dim)
        self._input = np.zeros(self._input_dim)

    @abstractmethod
    def step(self):
        """Increment the noise step for time dependent noise classes."""
        pass

    def set_input(self, u: Array):
        """Sets the input of the dynamical system. Assuming this class gets called multiple times between controller calls. We therefore store the input as a constant in the class.

        Args:
            u: Input to the dynamical system.
        """
        # TODO check for shape?
        self.data = self.data.replace(u=u)


class KalmanFilter(Estimator):
    """Unscented Kalman Filter class that wraps the filterPY toolbox."""

    def __init__(
        self,
        dt: float,
        model: str = "so_rpy",
        config: str = "cf2x_L250",
        filter_type: str = "UKF",
        estimate_rotor_vel: bool = False,
        estimate_dist_f: bool = False,
        estimate_dist_t: bool = False,
        initial_obs: dict[str, Array] | None = None,
    ):  # TODO give obs and info # analytical_mel_att
        """Initialize basic parameters.

        Args:
            dt: Time step between callings.
            model: The name of the model that is to be used.
            config: The setup configuration of the drone.
            filter_type: Either EKF or UKF
            estimate_rotor_vel: If the rotor speeds should be estimated, defaults to False.
            estimate_dist_f: If the disturbance forces should be estimated, defaults to False.
            estimate_dist_t: If the disturbance torques should be estimated, defaults to False.
            initial_obs: Optional, the initial observation of the environment's state. See the environment's observation space for details.
        """
        fx = dynamics_function(model, config)
        hx = observation_function
        # fx = jax.jit(dynamics_numeric(model, config))
        # hx = jax.jit(observation_function)

        dim_x = 13
        if estimate_rotor_vel:
            dim_x += 4
        if estimate_dist_f:
            dim_x += 3
        if estimate_dist_t:
            dim_x += 3
        dim_u = 4  # TODO from available models
        dim_z = 7  # pos, quat

        self.data = UKFData.create_empty(
            rotor_vel=estimate_rotor_vel,
            dist_f=estimate_dist_f,
            dist_t=estimate_dist_t,
            dim_u=dim_u,
            dim_z=dim_z,
        )

        dim_x = UKFData.get_state_dim(self.data)
        # print(f"dim_x={dim_x}")

        Q, R = self.create_covariance_matrices(
            dim_x=dim_x,
            dim_z=dim_z,
            varQ_pos=1e-6,
            varQ_quat=1e-2,
            varQ_rotor_vel=1e-5,
            varR_pos=1e-9,
            varR_quat=5e-9,
            dt=dt,
        )
        # Q = self.create_Q(
        #     dim_x=dim_x, varQ_pos=1e-6, varQ_quat=1e-6, varQ_vel=1e-3, varQ_ang_vel=1e-2, dt=dt
        # )
        # R = self.create_R(dim_z=dim_z, varR_pos=1e-8, varR_quat=3e-6, dt=dt)

        sigma_settings = SigmaPointsSettings.create(n=dim_x, alpha=1e-3, beta=2.0, kappa=0.0)
        self.settings = UKFSettings.create(SPsettings=sigma_settings, Q=Q, R=R, fx=fx, hx=hx)

        # Initialize state and covariance
        if initial_obs is not None:
            self.set_state(initial_obs["pos"], initial_obs["quat"])

    def set_state(self, pos: Array, quat: Array):
        """Sets pos and quat of the state to the given values and reduces the covariance value.

        With aggressive Q and R values, this needs to be called in the beginning. Otherwise the estimator might fail.
        """
        self.data = self.data.replace(pos=pos, quat=quat)
        dim_x = UKFData.as_state_array(self.data).shape[0]
        self.data = self.data.replace(covariance=np.eye(dim_x) * 1e-6)

    def create_covariance_matrices(
        self,
        dim_x: int,
        dim_z: int,
        varQ_pos: float,
        varQ_quat: float,
        varQ_rotor_vel: float,
        varR_pos: float,
        varR_quat: float,
        dt: float,
        varQ_dist_f: float = 1e-11,
        varQ_dist_t: float = 1e-18,  # TODO dist torque very noisy
    ) -> tuple[Array, Array]:
        """Creates sophisticated covariance matrices based on a paper.

        We assume that the linear elements (pos, vel) are correlated
        and the rotational elements (quat, ang_vel) are.
        For more information consult the filterpy library.
        """
        ### Set process noise covariance (tunable). Uncertainty in the dynamics. High Q -> less trust in model
        Q = np.eye(dim_x)

        # TODO maybe add correlation with rotor_vel!

        # Linear equations
        Q_xyz = Q_discrete_white_noise(
            dim=2, dt=dt, var=varQ_pos, block_size=3, order_by_dim=False
        )  # pos & vel
        Q[0:3, 0:3] = Q_xyz[0:3, 0:3]  # pos
        Q[7:10, 7:10] = Q_xyz[3:6, 3:6]  # vel
        Q[0:3, 7:10] = Q_xyz[0:3, 3:6]  # pos <-> vel
        Q[7:10, 0:3] = Q_xyz[3:6, 0:3]  # pos <-> vel

        # Rotational equations
        # since quat and ang_vel have different length, it's a little more tricky
        Q_rot = Q_discrete_white_noise(
            dim=2, dt=dt, var=varQ_quat, block_size=3, order_by_dim=False
        )  # quat & ang_vel
        Q[3:5, 3:5] = Q_rot[0:2, 0:2]  # quat12
        Q[5:7, 5:7] = Q_rot[0:2, 0:2]  # quat34
        Q[10:13, 10:13] = Q_rot[3:6, 3:6]  # ang_vel
        # We know that all quat_dot dependent on all ang_vel
        # => fill in the whole 4x3 and 3x4 matrix blocks with the variance
        Q[3:7, 10:13] = Q_rot[0, 3]  # quat <-> ang_vel
        Q[10:13, 3:7] = Q_rot[3, 0]  # quat <-> ang_vel

        i = 13  # for keeping how big the state (index) is

        # Motor forces TODO: maybe add correlation of acceleration (and angular acc) to motor forces
        if self.data.rotor_vel is not None:
            Q_forces = Q_discrete_white_noise(
                dim=2, dt=dt, var=varQ_rotor_vel, block_size=1, order_by_dim=False
            )  # Motor Forces
            Q[i : i + 2, i : i + 2] = np.eye(2) * Q_forces[0, 0]  # forces 1&2
            Q[i + 2 : i + 4, i + 2 : i + 4] = np.eye(2) * Q_forces[0, 0]  # forces 3&4
            Q[7:10, i : i + 4] = Q_forces[0, 1]  # forces <-> vel
            Q[i : i + 4, 7:10] = Q_forces[1, 0]  # forces <-> vel
            Q[10:13, i : i + 4] = (
                Q_forces[0, 1] * 0.04
            )  # forces <-> ang_vel, times arm length (F=rxl)
            Q[i : i + 4, 10:13] = (
                Q_forces[1, 0] * 0.04
            )  # forces <-> ang_vel, times arm length (F=rxl)
            # Q[i : i + 4] *= varQ_rotor_vel  # TODO move index as in dataclass example and make optional
            i = i + 4

        # External forces and torques TODO Maybe make x and v dependend on F uncertainty and same for torque
        if self.data.dist_f is not None:  # TODO move index as in dataclass example
            Q[i : i + 3, i : i + 3] *= varQ_dist_f  # Force
            i = i + 3
        if self.data.dist_t is not None:
            Q[i : i + 3, i : i + 3] *= varQ_dist_t  # Torque

        ### Set measurement noise covariance (tunable). Uncertaints in the measurements. High R -> less trust in measurements
        R = np.eye(dim_z)  # Assuming uncorrelated noise
        # very low noise on the position ("mm precision" => even less noise)
        R[:3, :3] = R[:3, :3] * varR_pos
        # "high" measurements noise on the angles, estimate: 0.01 constains all values => std=3e-3 TODO look at new quat measurements
        R[3:, 3:] = R[3:, 3:] * varR_quat

        return Q, R

    def create_Q(
        self,
        dim_x: int,
        varQ_pos: float,
        varQ_quat: float,
        varQ_vel: float,
        varQ_ang_vel: float,
        dt: float,
        varQ_rotor_vel: float = 1e-1,
        varQ_dist_f: float = 1e-9,
        varQ_dist_t: float = 1e-12,
    ) -> Array:
        """TODO."""
        Q = np.eye(dim_x)
        # TODO remove, just for testing
        Q[0:3] *= varQ_pos
        Q[3:7] *= varQ_quat
        Q[7:10] *= varQ_vel
        Q[10:13] *= varQ_ang_vel
        i = 13
        if self.data.rotor_vel is not None:
            Q[i : i + 4] *= varQ_rotor_vel
            i = i + 4
        if self.data.dist_f is not None:
            Q[i : i + 3, i : i + 3] *= varQ_dist_f  # Force
            i = i + 3
        if self.data.dist_t is not None:
            Q[i : i + 3, i : i + 3] *= varQ_dist_t  # Torque
        return Q

    def create_R(self, dim_z: int, varR_pos: float, varR_quat: float, dt: float) -> Array:
        """TODO."""
        ### Set measurement noise covariance (tunable). Uncertaints in the measurements. High R -> less trust in measurements
        R = np.eye(dim_z)  # Assuming uncorrelated noise
        # very low noise on the position ("mm precision" => even less noise)
        R[:3, :3] = R[:3, :3] * varR_pos
        # "high" measurements noise on the angles, estimate: 0.01 constains all values => std=3e-3 TODO look at new quat measurements
        R[3:, 3:] = R[3:, 3:] * varR_quat
        return R

    def step(self, pos: Array, quat: Array, dt: float, command: Array | None = None) -> UKFData:
        """Steps the UKF by one. Doing one prediction and correction step.

        Args:
            pos: Latest observation of the position
            quat: Latest observation of the quaternion
            dt: Optional, time step size. If not specified, default time is used
            command: Optional, latest input to the system

        Return:
            New state prediction
        """
        # Check if time step is positive
        if dt <= 0:
            return self.data

        # Update the input
        if command is not None:
            self.set_input(command)

        # Update observation and dt
        # dt hast to be vectorized to work properly in jax
        self.data = self.data.replace(z=np.concat((pos, quat)), dt=np.array([dt]))

        # if self.data.dt > 0:  # TODO make dt check more elegant and catch all errors
        # self.data = ukf_predict(self.data, self.settings)
        # self.data = ukf_predict(self.data, self.settings)
        self.data = ukf_predict_correct(self.data, self.settings)

        return self.data

    def predict(self, dt: float, command: Array | None = None) -> UKFData:
        """TODO."""
        # Check if time step is positive
        if dt <= 0:
            return self.data

        # Update the input
        if command is not None:
            self.set_input(command)

        # Update observation and dt
        # dt hast to be vectorized to work properly in jax
        self.data = self.data.replace(dt=np.array([dt]))

        self.data = ukf_predict(self.data, self.settings)

        return self.data

    def correct(self, pos: Array, quat: Array, command: Array | None = None) -> UKFData:
        """TODO."""
        # Update the input
        if command is not None:
            self.set_input(command)

        # Update observation and dt
        # dt hast to be vectorized to work properly in jax
        self.data = self.data.replace(z=np.concat((pos, quat)))

        # print(f"{quat=}, {self.data.quat=}, {self.data.u[-2]=}")

        self.data = ukf_correct(self.data, self.settings)

        return self.data


# test = KalmanFilter(1.0 / 200)

# times = []

# for i in range(1000):
#     print(f"iteration {i}")
#     t1 = time.perf_counter()
#     test.step(np.array([0, 0, 0, 0, 0, 0, 1]))
#     t2 = time.perf_counter()
#     times.append(t2 - t1)

# print(f"Avg = {np.mean(times) * 1000}ms")



"""The ROS connector module provides an interface for communication with ROS2 topics.

It defines the ROSConnector class, which handles communication with ROS2 through multiprocessing
to ensure non-blocking operation. The module is responsible for:

* Providing real-time state estimation (position, orientation, velocity, ...) for drones
* Tracking objects (gates, obstacles) through ROS /tf
* Publishing commands to ROS topics asynchronously for estimator nodes and logging

The main objective for the ROSConnector is to interface with ROS2 messages with minimal latency. It
uses a multiprocessing architecture to prevent ROS callbacks from blocking the main application
thread. Data is shared between processes using synchronized shared memory arrays, allowing for
efficient, thread-safe data access without copying.
"""

from __future__ import annotations

import atexit
import multiprocessing as mp
import time
from functools import partial
from queue import Empty
from typing import TYPE_CHECKING
from uuid import uuid4

import numpy as np
import rclpy
from geometry_msgs.msg import PoseStamped, TwistStamped, WrenchStamped
from rclpy.qos import QoSHistoryPolicy, QoSProfile, QoSReliabilityPolicy
from std_msgs.msg import Float64MultiArray
from tf2_msgs.msg import TFMessage

if TYPE_CHECKING:
    from multiprocessing.sharedctypes import Synchronized, SynchronizedArray
    from multiprocessing.synchronize import Event

    from numpy.typing import NDArray


class ROSConnector:
    """A non-blocking interface for ROS2 communication using multiprocessing.

    The ROSConnector class provides a thread-safe way to receive data from ROS2 topics and publish
    commands without blocking the main application thread. It uses separate processes for
    subscribing to topics and publishing commands, with data shared through synchronized memory
    arrays.

    The class supports two types of tracking:

    * Simple pose tracking from `/tf` topics
    * Full state estimation including velocity, angular velocity, and disturbance forces from an
      estimator node.

    Data is accessed through property methods that convert the shared memory arrays into
    dictionaries keyed by object names.

    Example:
        >>> # Track drone position and state from estimator
        >>> conn = ROSConnector(estimator_names=["cf1"], cmd_topic="/drones/cf1/cmd", timeout=1.0)
        >>> # Access the latest position data
        >>> drone_pos = conn.pos["cf1"]
        >>> # Publish a command (non-blocking)
        >>> conn.publish_cmd(np.array([1.0, 0.0, 0.0, 0.0]))
        >>> # Clean up when done
        >>> conn.close()
    """

    def __init__(
        self,
        tf_names: list[str] | None = None,
        estimator_names: list[str] | None = None,
        cmd_topic: str | None = None,
        timeout: float = 1.0,
        outdated_after: float = 0.1,
    ):
        """Initialize the ROSConnector with the specified tracking objects and command topic.

        Note:
            It is strongly recommended to use the timeout parameter to ensure that the connector has
            valid data for all objects. Otherwise, data may contain NaN values until the first
            update has been received.

        Args:
            tf_names: Names of objects to track using only pose data from `/tf` topic. If None,
                nothing will be tracked.
            estimator_names: Names of objects to track with full state estimation. If None, nothing
                will be tracked.
            cmd_topic: Topic name for publishing commands. If None, command publishing is disabled.
            timeout: If greater than 0, wait for position updates of all tracked objects before
                returning.
            outdated_after: Time after which measurements are considered outdated and property
                is_outdated will return true

        Raises:
            AssertionError: If ROS is not initialized before creating the connector.
            TimeoutError: If updates for all tracked objects aren't received within the specified
                timeout period.
        """
        assert rclpy.ok(), "ROS must be initialized before creating a ROSConnector instance."
        self.tf_names = [] if tf_names is None else tf_names
        self.estimator_names = [] if estimator_names is None else estimator_names
        assert not set(self.tf_names).intersection(set(self.estimator_names)), (
            "Duplicate names in tf and estimator"
        )
        assert len(self.tf_names) == len(set(self.tf_names)), "Duplicate items in tf_names"
        assert len(self.estimator_names) == len(set(self.estimator_names)), (
            "Duplicate items in estimator_names"
        )
        self.names = self.tf_names + self.estimator_names
        assert outdated_after > 0, f"outdated_after needs to be > 0, was {outdated_after}"
        self._outdated_after = outdated_after
        # Create synchronized, shared arrays for all quantities
        ctx = mp.get_context("spawn")
        self._pos = ctx.Array("f", [float("nan")] * len(self.names) * 3)
        self._quat = ctx.Array("f", [float("nan")] * len(self.names) * 4)
        self._vel = ctx.Array("f", [float("nan")] * len(self.estimator_names) * 3)
        self._ang_vel = ctx.Array("f", [float("nan")] * len(self.estimator_names) * 3)
        self._disturbance = ctx.Array("f", [float("nan")] * len(self.estimator_names) * 6)
        self._forces = ctx.Array("f", [float("nan")] * len(self.estimator_names) * 4)
        self._times = ctx.Array("d", [float("nan")] * len(self.names))
        # Note that times needs to be double, since time.time() might be a large number

        self.shutdown = ctx.Event()
        atexit.register(lambda: self.shutdown.set())  # Ensure processes are killed on exit
        self.processes = []
        if self.tf_names:  # Create a process for the /tf callback
            tf_process = ctx.Process(
                target=tf_update,
                args=(self.tf_names, self._pos, self._quat, self._times, self.shutdown),
            )
            self.processes.append(tf_process)

        if self.estimator_names:  # Create a process for each estimator callback
            estimator_process = ctx.Process(
                target=estimate_update,
                args=(
                    self.names,
                    self.estimator_names,
                    self._pos,
                    self._quat,
                    self._vel,
                    self._ang_vel,
                    self._disturbance,
                    self._forces,
                    self._times,
                    self.shutdown,
                ),
            )
            self.processes.append(estimator_process)

        self.cmd_pub, self.cmd_queue = None, None
        if cmd_topic is not None:  # Create a process for the command publisher
            self.cmd_queue = ctx.Queue(maxsize=10)
            self.cmd_pub = ctx.Process(
                target=command_publisher, args=(self.cmd_queue, cmd_topic, self.shutdown)
            )
            self.cmd_pub.start()

        for p in self.processes:
            p.start()

        # Timeouts are highly recommended to avoid reading NaNs and crashing while processes are
        # still spinning up
        if timeout:
            tstart = time.time()
            while time.time() - tstart < timeout:
                if self.active:
                    break
                time.sleep(0.01)  # Processes are spinning, so we can sleep here
            else:
                self.shutdown.set()
                missing_objects = [name for name, pos in self.pos.items() if np.any(np.isnan(pos))]
                tf_objects = [n for n in missing_objects if n in self.tf_names]
                estimator_objects = [n for n in missing_objects if n in self.estimator_names]
                missing_msg = ""
                if any(tf_objects):
                    missing_msg += (
                        f"\nTF data is missing for {tf_objects}. Is ros2 publishing tf poses?"
                    )
                if any(estimator_objects):
                    missing_msg += (
                        f"\nEstimator data is missing for {estimator_objects} objects. Is the "
                        "estimator node running?"
                    )
                raise TimeoutError(
                    f"Timeout while fetching initial positions for tracked objects. {missing_msg}"
                )

    @property
    def pos(self) -> dict[str, NDArray]:
        """The latest position data for all tracked objects."""
        pos = np.asarray(self._pos, dtype=np.float32).reshape(-1, 3)
        return {n: pos[i] for i, n in enumerate(self.names)}

    @property
    def quat(self) -> dict[str, NDArray]:
        """The latest orientation data for all tracked objects."""
        quat = np.asarray(self._quat, dtype=np.float32).reshape(-1, 4)
        return {n: quat[i] for i, n in enumerate(self.names)}

    @property
    def vel(self) -> dict[str, NDArray]:
        """The latest velocity data for all tracked objects."""
        vel = np.asarray(self._vel, dtype=np.float32).reshape(-1, 3)
        return {n: vel[i] for i, n in enumerate(self.estimator_names)}

    @property
    def ang_vel(self) -> dict[str, NDArray]:
        """The latest angular velocity data for all tracked objects."""
        ang_vel = np.asarray(self._ang_vel, dtype=np.float32).reshape(-1, 3)
        return {n: ang_vel[i] for i, n in enumerate(self.estimator_names)}

    @property
    def disturbance(self) -> dict[str, NDArray]:
        """The latest disturbance data for all tracked objects."""
        disturbance = np.asarray(self._disturbance, dtype=np.float32).reshape(-1, 6)
        return {n: disturbance[i] for i, n in enumerate(self.estimator_names)}

    @property
    def forces(self) -> dict[str, NDArray]:
        """The latest estimated forces data for all tracked objects."""
        forces = np.asarray(self._forces, dtype=np.float32).reshape(-1, 4)
        return {n: forces[i] for i, n in enumerate(self.estimator_names)}

    @property
    def times(self) -> dict[str, np.float32]:
        """The times of the last received data for all tracked objects."""
        times = np.asarray(self._times, dtype=np.float64).reshape(-1)
        return {n: times[i] for i, n in enumerate(self.estimator_names)}

    @property
    def is_outdated(self) -> dict[str, np.bool]:
        """The latest estimated forces data for all tracked objects."""
        times = np.asarray(self._times, dtype=np.float64).reshape(-1)
        is_outdated = times < time.time() - self._outdated_after
        return {n: is_outdated[i] for i, n in enumerate(self.estimator_names)}

    @property
    def active(self) -> bool:
        """Check if each object has received an update already."""
        # Check if drone is being tracked and if drone has already received updates
        pos_ready = not np.any(np.isnan(np.array(self._pos, dtype=np.float32)))
        quat_ready = not np.any(np.isnan(np.array(self._quat, dtype=np.float32)))
        vel_ready = not np.any(np.isnan(np.array(self._vel, dtype=np.float32)))
        ang_vel_ready = not np.any(np.isnan(np.array(self._ang_vel, dtype=np.float32)))
        return pos_ready and quat_ready and vel_ready and ang_vel_ready

    def publish_cmd(self, value: NDArray):
        """Publish a command to ROS.

        This is a non-blocking function that puts the command into a multiprocessing queue. Commands
        are picked up by the command publisher process and published to the ROS topic
        asynchronously.

        Note:
            The command publisher topic must be set on initialization. Otherwise this function will
            raise an error.
        """
        if self.cmd_pub is None:
            raise ValueError("Command publisher not initialized.")
        self.cmd_queue.put(value)

    def close(self):
        """Unregister the ROS subscriptions and shut down the node."""
        self.shutdown.set()
        for p in self.processes:
            p.join()


def tf_update(
    names: list[str],
    pos: list[SynchronizedArray],
    quat: list[SynchronizedArray],
    times: list[Synchronized],
    shutdown: Event,
):
    """Update the shared memory arrays with the latest position and orientation data from /tf.

    Args:
        names: The names of the objects to track.
        pos: The shared memory arrays for the position data.
        quat: The shared memory arrays for the orientation data.
        times: The shared memory arrays for the time data.
        shutdown: The event to signal shutdown.
    """
    rclpy.init()
    node = rclpy.create_node("tf_updater_" + uuid4().hex)
    fn = partial(tf_callback, names=names, pos=pos, quat=quat, times=times)
    sub = node.create_subscription(TFMessage, "/tf", fn, 10)
    while not shutdown.is_set():
        rclpy.spin_once(node, timeout_sec=0.1)
    sub.destroy()
    node.destroy_node()


def estimate_update(
    all_names: list[str],
    names: list[str],
    pos: list[SynchronizedArray],
    quat: list[SynchronizedArray],
    vel: list[SynchronizedArray],
    ang_vel: list[SynchronizedArray],
    disturbance: list[SynchronizedArray],
    forces: list[SynchronizedArray],
    times: list[Synchronized],
    shutdown: Event,
):
    """Update the shared memory arrays with the latest pose data from the estimator node.

    Args:
        all_names: The names of all objects that are being tracked.
        names: The names of the objects only being tracked by the estimator.
        pos: The shared memory arrays for the position data.
        quat: The shared memory arrays for the orientation data.
        vel: The shared memory arrays for the velocity data.
        ang_vel: The shared memory arrays for the angular velocity data.
        disturbance: The shared memory arrays for the disturbance data.
        forces: The shared memory arrays for the forces data.
        times: The shared memory arrays for the time data.
        shutdown: The event to signal shutdown.
    """
    rclpy.init()
    node = rclpy.create_node("estimate_updater_" + uuid4().hex)

    qos_profile = QoSProfile(
        reliability=QoSReliabilityPolicy.BEST_EFFORT, history=QoSHistoryPolicy.KEEP_LAST, depth=1
    )

    subs = []
    for name in names:
        # Pose (pos and quat) are shared with all_names, so the index changes!
        fn = partial(pose_callback, idx=all_names.index(name), pos=pos, quat=quat, times=times)
        sub = node.create_subscription(
            PoseStamped, f"/drones/{name}/estimate/pose", fn, qos_profile
        )
        subs.append(sub)
        fn = partial(twist_callback, idx=names.index(name), vel=vel, ang_vel=ang_vel)
        sub = node.create_subscription(
            TwistStamped, f"/drones/{name}/estimate/twist", fn, qos_profile
        )
        subs.append(sub)
        fn = partial(disturbance_callback, idx=names.index(name), disturbance=disturbance)
        sub = node.create_subscription(
            WrenchStamped, f"/drones/{name}/estimate/wrench", fn, qos_profile
        )
        subs.append(sub)
        fn = partial(forces_callback, idx=names.index(name), forces=forces)
        sub = node.create_subscription(
            Float64MultiArray, f"/drones/{name}/estimate/forces", fn, qos_profile
        )
        subs.append(sub)

    while not shutdown.is_set():
        rclpy.spin_once(node, timeout_sec=0.1)
    for sub in subs:
        sub.destroy()
    node.destroy_node()


def command_publisher(queue: mp.Queue, cmd_topic: str, shutdown: Event):
    """Publish commands from a queue to a ROS topic.

    Args:
        queue: The queue containing commands to publish.
        cmd_topic: The topic name.
        shutdown: The event to signal shutdown.
    """
    rclpy.init()
    node = rclpy.create_node("cmd_publisher_" + uuid4().hex)
    publisher = node.create_publisher(Float64MultiArray, cmd_topic, 10)

    while not shutdown.is_set():
        try:
            cmd = None
            cmd = queue.get(timeout=0.1)  # Get the latest message from the queue. Block if empty
            while True:  # Get the latest message if queue holds more
                cmd = queue.get_nowait()
        except Empty:  # We have cleared the queue or timed out, either is fine
            pass
        if cmd is not None:
            cmd = np.array(cmd, dtype=np.float64)
            publisher.publish(Float64MultiArray(data=cmd))

    node.destroy_node()


def tf_callback(
    data: TFMessage,
    names: list[str],
    pos: SynchronizedArray,
    quat: SynchronizedArray,
    times: SynchronizedArray,
):
    """Save the pose data from the /tf topic into shared memory arrays.

    Args:
        data: The TFMessage message.
        names: The names of the objects to track.
        pos: The shared memory array for the position values.
        quat: The shared memory array for the orientation values.
        times: The shared memory array for the time values.
    """
    for tf in data.transforms:
        name = tf.child_frame_id.split("/")[-1]
        if name in names:
            T, Rot = tf.transform.translation, tf.transform.rotation
            idx = names.index(name)
            pos[idx * 3 : (idx + 1) * 3] = [T.x, T.y, T.z]
            quat[idx * 4 : (idx + 1) * 4] = [Rot.x, Rot.y, Rot.z, Rot.w]
            times[idx] = time.time()


def pose_callback(
    data: PoseStamped,
    idx: int,
    pos: SynchronizedArray,
    quat: SynchronizedArray,
    times: Synchronized,
):
    """Save the drone pose from the estimator node.

    Args:
        data: The PoseStamped message.
        idx: The index of the object the pose is referring to.
        pos: The shared memory array for the position values.
        quat: The shared memory array for the orientation values.
        times: The shared memory array for the time values.
    """
    T, Rot = data.pose.position, data.pose.orientation
    pos[idx * 3 : (idx + 1) * 3] = [T.x, T.y, T.z]
    quat[idx * 4 : (idx + 1) * 4] = [Rot.x, Rot.y, Rot.z, Rot.w]
    times[idx] = time.time()


def twist_callback(
    data: TwistStamped, idx: int, vel: SynchronizedArray, ang_vel: SynchronizedArray
):
    """Save the drone twist (velocity and angular velocity) from the estimator node.

    Args:
        data: The TwistStamped message.
        idx: The index of the object the velocity is referring to.
        vel: The shared memory array for the velocity values.
        ang_vel: The shared memory array for the angular velocity values.
    """
    linear, angular = data.twist.linear, data.twist.angular
    vel[idx * 3 : (idx + 1) * 3] = [linear.x, linear.y, linear.z]
    ang_vel[idx * 3 : (idx + 1) * 3] = [angular.x, angular.y, angular.z]


def disturbance_callback(data: WrenchStamped, idx: int, disturbance: SynchronizedArray):
    """Save the disturbance force and torque from the estimator node.

    Args:
        data: The WrenchStamped message.
        idx: The index of the object the force is referring to.
        disturbance: The shared memory array for the disturbances.
    """
    force, torque = data.wrench.force, data.wrench.torque
    disturbance[idx * 6 : (idx + 1) * 6] = [force.x, force.y, force.z, torque.x, torque.y, torque.z]


def forces_callback(data: Float64MultiArray, idx: int, forces: SynchronizedArray):
    """Save the estimated forces from the estimator node.

    Args:
        data: The Float64MultiArray message.
        idx: The index of the object the force is referring to.
        forces: The shared memory array for the forces.
    """
    forces[idx * 4 : (idx + 1) * 4] = data.data
